#############################
# THINGS WE'VE ALREADY DONE #
#############################
#import data
train.data <- read.csv("data/titanic/train.csv", stringsAsFactors = F)
test.data <- read.csv("data/titanic/test.csv", stringsAsFactors = F)

# we add the Survived column to the test dataset so that it has the same structure as the
# training set and the two can be merged
test.data$Survived <- rep(x = NA, times = length(test.data))

# we merge the two datasets so that all data preparation processing is applied 
# to both training and test datasets
merged.data <- rbind(train.data, test.data)

# transform the Survived variable into a factor variable (like Enum in Java)
merged.data$Survived <- factor(x = merged.data$Survived, levels = c(0,1), labels = c("No", "Yes"))

# transform the Pclass variable into a factor
merged.data$Pclass <- factor(x = merged.data$Pclass, levels = c(1,2,3), 
                             labels = c("1st", "2nd", "3rd"), ordered = TRUE)

# transform the Sex variable into a factor variable 
merged.data$Sex <- factor(merged.data$Sex)

# transform the Embarked variable into a factor variable 
merged.data$Embarked <- factor(merged.data$Embarked)
# the Embarked variable has 2 empty string values; replace these
# with the most dominant (majority) value
table(merged.data$Embarked) # the majority value is "S"
merged.data$Embarked[merged.data$Embarked == ""] <- "S"
merged.data$Embarked <- factor(merged.data$Embarked)

# replace the missing Cabin values for 1st class passangers with NAs 
merged.data[ (merged.data$Cabin == "") & (merged.data$Pclass == 1), "Cabin"] <- NA

# replace the missing Fare value with the mean value for the Fare variable of the 
# passanges of the same class
missing.fare.index <- which( is.na(merged.data$Fare) )
missing.fare.pclass <- merged.data$Pclass[missing.fare.index]
mean.fare <- mean(x = merged.data$Fare[merged.data$Pclass == missing.fare.pclass], na.rm = T)
merged.data$Fare[which(is.na(merged.data$Fare))] <- mean.fare


# extract passangers' titles from the values of the Name variable:
merged.data$Title <- sapply(X = merged.data$Name, 
                            FUN = function(x) { strsplit(x, "[,.]")[[1]][2] })

# remove the leading blank space from of the titles
merged.data$Title <- sapply(X = merged.data$Title, 
                            FUN = function(x) { substr(x, start = 2, stop = nchar(x))} )

# aggregate the extracted titles into broader categories
ladies <- c("Dona", "Lady", "Mme", "Mrs", "the Countess")
merged.data$Title[ merged.data$Title %in% ladies ] <- "Mrs" # the %in% operator checks to see if a value is part of the given vector 

misses = c("Miss", "Mlle")
merged.data$Title[ merged.data$Title %in% misses ] <- "Ms"

gentelman <- c("Capt", "Col", "Don", "Dr", "Major", "Mr", "Rev", "Sir")
merged.data$Title[ merged.data$Title %in% gentelman ] <- "Mr"

masters <- c("Master", "Jonkheer")
merged.data[ merged.data$Title %in% masters, "Title"] <- "Master"

# if Age is avaialble and greater than 18, and the Title is 'Ms', change the title to 'Mrs'
merged.data$Title[(merged.data$Age > 18) & merged.data$Title == "Ms"] <- "Mrs"
# if Age is avaialble and less than 18, and the Title is 'Mr', change the title to 'Master'
merged.data$Title[(merged.data$Age < 18) & (merged.data$Title == "Mr")] <- "Master"

# transform Title into a factor variable
merged.data$Title <- factor(merged.data$Title)

# introducing a feature representing the number of family members one is travelling with
merged.data$FamilySize <- merged.data$SibSp + merged.data$Parch
# aggregating instances with 3+ family members:
merged.data$FamilySize[ merged.data$FamilySize > 3 ] <- 3
# and turning FamilySize into a factor
merged.data$FamilySize <- factor(x = merged.data$FamilySize, levels = c(0,1,2,3), 
                                 labels = c("Alone", "1", "2", "3+"))

################################
######  MODEL BUILDING    ######
################################

# since our test dataset does not have values for the output variable (Survived),
# we will partion the training part of our merged dataset into two parts: 
# training and validation 
## <!explain the purpose of validation!>
# once we achieve sufficiently good prediction results on the validation set, 
# we can do the prediction on the test set and submit our results to the Kaggle website

# we want to randomly select instances for the training and the validation set;
# whenever working with random processes in R, we should first set the seed for 
# the random process so that we can replicate our results
set.seed(1234)
# indices of all the instances with known outcome variable
indices.all <- 1:891 
# we randomly select 75% of all the indices for training and the rest for validation
indices.train <- sample(x = indices.all, 
                        size = round( length(indices.all)*0.75), 
                        replace = FALSE)
indices.train[1:10]
indices.validate <- indices.all [ !(indices.all %in% indices.train) ]
# using the indices vectors, we can create datasets for training and validation 
train <- merged.data[ indices.train, ]
validate <- merged.data[ indices.validate, ]


# We'll create a prediction model using Decision Trees
# Hence, we need to load the rpart R package for working with decision trees
# note: this is just one of the available R packages for working with decision trees
install.packages('rpart')
library(rpart)

# we fit the model using the rpart function
# by setting the value "class" of the method parameter we specify that we want to 
# use the majority vote principle for the class prediction
tree.model <- rpart(Survived ~ Sex + Title + FamilySize + Pclass + Embarked + Fare, 
                    data = train, method = "class")
# print the model
print(tree.model)

# let's plot the tree, to understand it better
library(rpart.plot)
par(mfrow = c(1,1))
prp(tree.model, type = 3, extra = 1)

# we can draw a more fancy plot of our tree model
# to do that, we will install some additional packages
library(rattle)
library(RColorBrewer)
fancyRpartPlot(tree.model)

# let's verify our model on the validation set
tree.predict <- predict(object = tree.model, newdata = validate, type = "class")
# see what the predictions look like
tree.predict[1:10]
table(tree.predict)
# we can create the confusion matrix 
## <!explain confusion matrix and the evaluation metrics!>
cm <- as.matrix ( table( true = validate$Survived, predicted = tree.predict ) )
cm

# since we'll need to compute evaluation metrics couple of times, it's handy to have
# a function for that; the f. receives the confusion matrix and the number of observations,
# and returns a named vector with the values for accuracy, precision and recall
compute.eval.measures <- function(cm, n) {
  acc <- (cm[1,1] + cm[2,2])/n
  
  # since we're interested in predicting survival, we'll compute precision and recall
  # using values of the confusion matrix that are related to the Yes outcome
  prec <- cm[2,2] / (cm[1,2] + cm[2,2]) 
  rec <- cm[2,2] / (cm[2,1] + cm[2,2])
  f1 <- (2*prec*rec)/(prec+rec)
  # return the performance metrics
  c(accuracy = acc, precision = prec, recall = rec, F1 = f1)  
}

tree.model.eval <- compute.eval.measures(cm, length( tree.predict ))
tree.model.eval

# the rpart function uses a number of parameters to control the growth of a tree
# (in the above call of the rpart f. we relied on the default values of those parameters)
# to inspect the parameters and their defaults, type: 
?rpart.control
# let's now change some of these parameters to grow a deeper/larger tree
# for instance, we can change 
# - cp - it stops splits that aren’t deemed important enough (i.e., the splits that do not 
# increase the fittness of the model by at least a factor of cp)
# - minsplit - minimum number of instances in a node for a split to be attempted at that node
tree.model2 <- rpart(Survived ~ Sex + Title + FamilySize + Pclass + Embarked + Fare, 
                     data = train, method = "class", 
                     control=rpart.control(minsplit=15, cp=0.0025))

print(tree.model2)
prp(tree.model2, type = 3, extra = 1)
# the following command - for drawing the fancy tree plot - tend to crush R, 
# so, think twice before calling it 
# fancyRpartPlot(tree.model2)

# let's measure the model's performance on the training set
train.predict <- predict(object = tree.model2, newdata = train, type = "class")
train.cm <- as.matrix ( table( true = train$Survived, predicted = train.predict ) )
train.metrics <- compute.eval.measures(train.cm, length( train.predict ))

# let's now check the same metrics on the validation set
validate.predict <- predict(object = tree.model2, newdata = validate, type = "class")
validate.cm <- as.matrix ( table( true = validate$Survived, predicted = validate.predict ) )
validate.metrics <- compute.eval.measures(validate.cm, length( validate.predict ))

# we can create a data frame of the computed metrics to have a better view of the 
# metrics and how they differ between the training and the validation set
tree.model2.eval <- data.frame( rbind(train.metrics, validate.metrics), 
                                row.names = c("train", "validate") )
names(tree.model2.eval) <- c("accuracy", "precision", "recall", "F1")
tree.model2.eval
# obviously by growing complex trees we run the risk of overfitting the model to the training set
## <! explain over-fitting and under-fitting !>


# We'll now do the cross-validation to determine optimal CP, i.e., the CP value
# that results in the lowest error (or highest model accuracy) 
## <!explain cross-validation!>

# we'll now determine the optimal value for cp through cross-validation
# install cross-validation packages
install.packages("caret")
library(caret)
install.packages("e1071")
library(e1071)

# define cross-validation parameters
set.seed(2015)
numFolds = trainControl( method = "cv", number = 10 )
cpGrid = expand.grid( .cp = seq(0.0001,0.01,0.0005)) 
# perform the cross validation
train(Survived ~ Sex + Title + FamilySize + Pclass + Embarked + Fare, 
      data = train, method = "rpart", 
      trControl = numFolds, tuneGrid = cpGrid )

# as the optimal CP value, we'll choose the one that is associated with both
# low error (xerror) and not-overly complex tree (i.e., number of splits is not overly large)
# after we choose the optimal value for the CP parameter, we can now prune the tree 
# to improve its performance
pruned.tree.model2 <- prune(tree = tree.model2, cp = 0.0051)

prp(pruned.tree.model2, type = 3, extra = 1)
# the following command - for drawing the fancy tree plot - tend to crush R, 
# so, think twice before calling it 
# fancyRpartPlot( pruned.tree.model2 )

# let's check how it performs on the validation set
val.pred.pruned <- predict(object = pruned.tree.model2, newdata = validate, type = "class")
val.pruned.cm <- as.matrix ( table( true = validate$Survived, predicted = val.pred.pruned ) )
val.pruned.cm
val.pruned.metrics <- compute.eval.measures(val.pruned.cm, length( validate.predict ))
val.pruned.metrics
