#############################
# THINGS WE'VE ALREADY DONE #
#############################
#import data
train.data <- read.csv("data/titanic/train.csv", stringsAsFactors = F)
test.data <- read.csv("data/titanic/test.csv", stringsAsFactors = F)

# we add the Survived column to the test dataset so that it has the same structure as the
# training set and the two can be merged
test.data$Survived <- rep(x = NA, times = length(test.data))

# we merge the two datasets so that all data preparation processing is applied 
# to both training and test datasets
merged.data <- rbind(train.data, test.data)

# transform the Survived variable into a factor variable (like Enum in Java)
merged.data$Survived <- factor(x = merged.data$Survived, levels = c(0,1), labels = c("No", "Yes"))

# transform the Pclass variable into a factor
merged.data$Pclass <- factor(x = merged.data$Pclass, levels = c(1,2,3), 
                             labels = c("1st", "2nd", "3rd"), ordered = TRUE)

# transform the Sex variable into a factor variable 
merged.data$Sex <- factor(merged.data$Sex)

# transform the Embarked variable into a factor variable 
merged.data$Embarked <- factor(merged.data$Embarked)

# replace the missing Cabin values for 1st class passangers with NAs 
merged.data[ (merged.data$Cabin == "") & (merged.data$Pclass == 1), "Cabin"] <- NA

# replace the missing Fare value with the mean value for the Fare variable of the 
# passanges of the same class
missing.fare.index <- which( is.na(merged.data$Fare) )
missing.fare.pclass <- merged.data$Pclass[missing.fare.index]
mean.fare <- mean(x = merged.data$Fare[merged.data$Pclass == missing.fare.pclass], na.rm = T)
merged.data$Fare[which(is.na(merged.data$Fare))] <- mean.fare


# extract passangers' titles from the values of the Name variable:
merged.data$Title <- sapply(X = merged.data$Name, 
                            FUN = function(x) { strsplit(x, "[,.]")[[1]][2] })

# remove the leading blank space from of the titles
merged.data$Title <- sapply(X = merged.data$Title, 
                            FUN = function(x) { substr(x, start = 2, stop = nchar(x))} )

# aggregate the extracted titles into broader categories
ladies <- c("Dona", "Lady", "Mme", "Mrs", "the Countess")
merged.data$Title[ merged.data$Title %in% ladies ] <- "Mrs" # the %in% operator checks to see if a value is part of the given vector 

misses = c("Miss", "Mlle")
merged.data$Title[ merged.data$Title %in% misses ] <- "Ms"

gentelman <- c("Capt", "Col", "Don", "Dr", "Major", "Mr", "Rev", "Sir")
merged.data$Title[ merged.data$Title %in% gentelman ] <- "Mr"

masters <- c("Master", "Jonkheer")
merged.data[ merged.data$Title %in% masters, "Title"] <- "Master"

# if Age is avaialble and greater than 18, and the Title is 'Ms', change the title to 'Mrs'
merged.data$Title[(merged.data$Age > 18) & merged.data$Title == "Ms"] <- "Mrs"
# if Age is avaialble and less than 18, and the Title is 'Mr', change the title to 'Master'
merged.data$Title[(merged.data$Age < 18) & (merged.data$Title == "Mr")] <- "Master"

# transform Title into a factor variable
merged.data$Title <- factor(merged.data$Title)

# introducing a feature representing the number of family members one is travelling with
merged.data$FamilySize <- merged.data$SibSp + merged.data$Parch
# aggregating instances with 3+ family members:
merged.data$FamilySize[ merged.data$FamilySize > 3 ] <- 3
# and turning FamilySize into a factor
merged.data$FamilySize <- factor(x = merged.data$FamilySize, levels = c(0,1,2,3), 
                                 labels = c("Alone", "1", "2", "3+"))


##########################################################
## ADDITIONAL DATA PREPARATION: IMPUTING MISSING VALUES ##
##########################################################

# Before applying the NB algorithm, let's first handle the missing values we've spotted
# previously (otherwise, those observations will be removed during the model building phase); 
# we've already handled the missing value of the Fare variable, now, we'll handle the others

# first, let's recall which variables have missing values
summary( merged.data )

# the Embarked variable has 2 missing values (in the dataset of 1309 observations);
# this makes it safe to replace the missing values with the dominant value ("majority label")
# let's first identify those missing values
which ( merged.data$Embarked == "" )
# and replace them with "S" as the majority label
merged.data$Embarked[ merged.data$Embarked == "" ] <- "S"
summary(merged.data$Embarked)
# there are no more empty strings, but empty string is kept as a factor value;
# we remove it by re-computing Embarked as a factor variable
merged.data$Embarked <- factor( merged.data$Embarked )

# we've already seen that the Age variable is very problematic in terms of missing values
# and therefore ignored it; let's now try to predict the missing Age values and use them for 
# building the overall model
summary( merged.data$Age )

# first, identify the observations with missing Age value
missing.indices <- which( is.na(merged.data$Age) )

# now, build a prediction model (namely, a regression tree)
# note that the method is "anova" as we are building now a regression (not classification) tree
# also note that we restricted the dataset to those instances that are not missing the Age value
# the minus sign (-) means 'all indices not present in the missing.indices vector'
require(rpart)
age.model <- rpart(Age ~ Pclass + Sex + Title + FamilySize, 
                   data = merged.data[ -missing.indices, ], method = "anova")
library(rpart.plot)
prp(age.model, type = 3, extra = 1)

# let's use the built model to predict the missing Age values and add them to our dataset
# first, predict Age value using the created model
missing.predicted <- predict(object = age.model, newdata = merged.data[missing.indices,] )
# add predicted values to the dataset
merged.data$Age[ missing.indices ] <- missing.predicted 
# check that now there are no missing Age values
summary( merged.data$Age )
# based on the summary statistics (mean, median, quantiles), the predicted values 
# haven't much affected the distribution of the Age variable  


##############################################################
## ADDITIONAL DATA PREPARATION: SCALLING AND DISCRETIZATION ##
##############################################################

# NB is often used with nominal variables; if numeric variables are to be used, they have to
# be either 1) represented as probabilities based on the appropriate probability distribution 
# (typically Normal dist.), or 2) discretized 
# In this case, we have two numeric variables - Age and Fare - and have to handle them in one
# of these two ways; however, before considering this, we should first scale them 

# Scaling attributes is very important as it prevents attributes with greater numeric ranges
# to dominate over those with smaller numeric ranges; it is recommended to scale each 
# numerical attribute to the range [−1, +1] or [0, 1]

# we will scale the two numeric attributes (Age and Fare) using the formula: 
# z = (x - min(x))/(max(x) - min(x))
merged.data$AgeScaled <- (merged.data$Age - min(merged.data$Age))/diff(range(merged.data$Age))
summary( merged.data$AgeScaled )
merged.data$FareScaled <- (merged.data$Fare - min(merged.data$Fare))/diff(range(merged.data$Fare))
summary( merged.data$FareScaled )

# now, let's check if these (scalled) numeric attributes are normally distributed
shapiro.test( merged.data$AgeScaled )
# p-value = 1.006e-12 indicates that the attribute is far from being normally dist.
shapiro.test( merged.data$FareScaled )
# the same applies to the Fare attribute (p-value < 2.2e-16)

# so, we'll follow the option of discretizing the two numeric attributes
install.packages('bnlearn')
library(bnlearn)

# select the two numeric attributes that are to be discretized
merged.numeric <- merged.data[, c("AgeScaled", "FareScaled")]
d <- discretize(data = merged.numeric, method = 'quantile', breaks = 10)
# we can check the bounderies of the 10 groups the AgeScaled values are split into
levels(d$AgeScaled)
# and see the distribution of the observations across the 10 groups
table(d$AgeScaled)
# we can do the same for the FareScaled attribute 
levels(d$FareScaled)
table(d$FareScaled)

# add these discretized variables to our dataset
merged.data$AgeScaledDiscr <- d$AgeScaled
merged.data$FareScaledDiscr <- d$FareScaled

#####################################################################
## SPLITTING DATA INTO TRAIN AND VALIDATION SETS (ALREADY COVERED) ##
#####################################################################

set.seed(135)
indices.all <- 1:891 # indices of all the instances with known outcome variable
# randomly select 75% of all the indices for training samples and the rest for validation
indices.train <- sample(x = indices.all, 
                        size = round( length(indices.all)*0.75), replace = FALSE)
indices.validate <- indices.all [ !(indices.all %in% indices.train) ]
# using the indices vectors, we can create datasets for training and validation 
train <- merged.data[ indices.train, ]
validate <- merged.data[ indices.validate, ]

################################
######  MODEL BUILDING    ######
################################

require(e1071)

nb.model <- naiveBayes(Survived ~ Pclass + Sex + Embarked + FamilySize + Title +
                         AgeScaledDiscr + FareScaledDiscr,
                       data = train, laplace = 1)
# note the 'laplace' parameter: it means that we want to use Laplace smoothing
# to prevent situation of getting 0 for probability estimate; in this case, we are 
# adding 1 to the numerator and k (=the number of possible attribute values) to the
# denominator of the probability equation for each categorical attribute

# now, let's do the prediction using the created model 
nb.pred <- predict(object = nb.model, newdata = validate, type = "class")

validate.cm <- as.matrix ( table( true = validate$Survived, predicted = nb.pred ) )
validate.cm

compute.eval.measures <- function(cm, n) {
  acc <- (cm[1,1] + cm[2,2])/n
  
  # since we're interested in predicting survival, we'll compute precision and recall
  # using values of the confusion matrix that are related to the Yes outcome
  prec <- cm[2,2] / (cm[1,2] + cm[2,2]) 
  rec <- cm[2,2] / (cm[2,1] + cm[2,2])
  f1 <- (2*prec*rec)/(prec+rec)
  TNR <- cm[1,1] / (cm[1,1] + cm[1,2])
  # return the performance metrics
  c(accuracy = acc, precision = prec, recall = rec, F1 = f1, sensitivity = rec, specificity = TNR)  
}

validate.metrics <- compute.eval.measures(validate.cm, length(nb.pred))
validate.metrics

# not that good results; 
# similar (even worse) results have been reported here: 
# http://sqldatamine.blogspot.com/2013/07/catastrophic-test-of-naive-bayes.html
# so, we should consider one (or more) of the following options:
# - exclude some of the used features
# - introduce new features
# - use another discretization method
# - choose another algorithm (NB might not be the best solution for this particular problem)


##########################
######  ROC CURVES  ######
##########################

# when doing class prediction based on a built prediction model, 
# we do not need to apply the "majority class" principle to do the classification 
# (i.e., use the 0.5 probability threshold to choose the class); 
# we can also do the classification based on the chosen threshold value
# e.g., we can set the probability threshold to 0.8 and say that we will consider 
# that a passanger survived if the predicted probability of survival is >=0.8

# let's now make predictions in such a way that we predict just the probabilities
# of the two possible values of the outcome (the Survived variable)  
nb.pred.prob <- predict(object = nb.model, newdata = validate, type = "raw")
# note that the type parameter is set to 'raw'
# let's inspect the first ten predictions, to see what they look like
nb.pred.prob[1:10,]

# the question is how to choose proper threshold value (i.e., the cutoff point)
# ROC curves can help with that
## <! explain ROC curves, AUC, sensitivity and specificity !>
# Sensitivity (the true positive rate) measures the proportion of positives which are 
# correctly identified as such; the same as recall
# Specificity (the true negative rate) measures the proportion of negatives which are 
# correctly identified as such


# to create ROC curves, we need to install and load the pROC package 
# http://web.expasy.org/pROC/
install.packages("pROC")
library(pROC)

nb.roc <- roc(response = validate$Survived, predictor = nb.pred.prob[,2] )
# note that nb.pred.prob[,2] means that we are taking the probabilities given in the 2nd col.
# of the matrix, that is, the probabilities of the 'Yes' value of the Survived variable
# since it is the positive class in the context of this problem

# we can check the Area Under the Curve (AUC) value 
nb.roc$auc
# it's rather good, as it is close to 1 (0.8717)

# we can also plot the ROC curve
plot.roc(validate$Survived, 
         nb.pred.prob[,2], 
         percent=T,  ci=T,  # compute conf. inter. (of AUC by default)  
         print.auc=T) # also print AUC and its conf. inter. on the plot 

# let's now use the ROC curve to determine the threshold; to that end, we'll use the coords f.
?coords
# see the documentation for the meaning of the parameters 'x' and 'best.method'
# with the 'ret' parameter we have choosen to return the threshold and accuracy values; 
# other metrics can be returned, as well (see the documentation)
nb.coords <- coords(roc = nb.roc, x = "best", best.method = "closest.topleft", 
                    ret = c("threshold", "accuracy", "sensitivity", "specificity"))
print(nb.coords)
threshold <- as.numeric( nb.coords[1] )

# let's now use the computed threshold to assign class labels to instances
nb.pred.labels <- sapply(nb.pred.prob[,2], 
                              function(x) { ifelse(test = (x >= threshold), yes = "Yes", no = "No")})
validate.cm2 <- as.matrix( table( true = validate$Survived, predicted = nb.pred.labels ) )
validate.cm2
compute.eval.measures( validate.cm2, length(nb.pred.labels) )
# accuracy and precision improved, precision even significantly improved
# recall got worse (as expected since precision increased significantly)


# We can also use ROC curves to (visually) compare different classifiers
# To illustrate that, let's compare the DT and NB classifiers
dt.model <- rpart(Survived ~ Sex + Title + FamilySize + Pclass + Embarked + Fare + Age, 
                    data = train, method = "class")
dt.pred.prob <- predict(object = dt.model, newdata = validate, type = "prob")

# plot the ROC curves
roc.dt <- plot.roc(validate$Survived, dt.pred.prob[,2], percent = T,  
                   main="Comparison of DT and NB classifiers", 
                   col="#1c61b6")  
roc.nb <- lines.roc(validate$Survived, nb.pred.prob[,2], percent = T, col="#008600")  
legend("bottomright", legend=c("DT", "NB"), col=c("#1c61b6", "#008600"), lwd=2)
# do a test to determine if there is a statisticaly significant difference between the 
# performance of the two classifiers
test.res <- roc.test(roc.dt, roc.nb)  
text(30, 90, labels=paste("p-value =", format.pval(test.res$p.value)), adj=c(0, .5))  
